/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef _GB_H
#define _GB_H

#include "types.h"
#include "rom.h"

#define ROM_FILENAME_SIZE 257

class gb_system
{
public:
   gb_system();
   ~gb_system();
   
   bool init();
   void reset();
   void mem_reset();
   void cpu_reset();
   
   char rom_filename[ROM_FILENAME_SIZE];
   bool load_rom(char* filename);
   bool loadrom_zip(char* filename);
   bool romloaded;
   void crdtype(byte value,byte romsize,byte ramsize);
   int process_rom_info(byte *rominfo);
   //CPU-----------------------------------------
   byte A;
   byte F; 
   word BC;
   word DE;
   word HL;
   word SP;   
   word PC;
   
   // flags in x86 order
   // size word because lahf loads flags to ah
   unsigned short flags; 

   int CFLAG;
   int HFLAG;
   int ZFLAG;
   int NFLAG;
   int system_type;
   
   #include "opcodes.h"
   
  /* void gbz80_inc(byte* reg);
   void gbz80_dec(byte* reg);
   void gbz80_rlca();
   void gbz80_rrca();
   void gbz80_rla();
   void gbz80_rra();
   void gbz80_rlc(byte* reg);
   void gbz80_rrc(byte* reg);
   void gbz80_rl(byte* reg);
   void gbz80_rr(byte* reg);
   void gbz80_sla(byte* reg);
   void gbz80_sra(byte* reg);
   void gbz80_srl(byte* reg);
   void gbz80_bit(byte reg,int num);
   void gbz80_swap(byte* reg);
   void gbz80_scf();
   void gbz80_ccf();
   void gbz80_cpl();
   void gbz80_and(byte reg);
   void gbz80_xor(byte reg);
   void gbz80_or(byte reg);
   void gbz80_cp(byte reg);
   void gbz80_add(byte reg);
   void gbz80_add16(unsigned short* reg, unsigned short reg2);
   void gbz80_addSP(unsigned short* reg, signed char reg2);
   void gbz80_adc(byte reg);
   void gbz80_sub(byte reg);
   void gbz80_sbc(byte reg);
   void gbz80_daa();
   void gbz80_das();*/
   
   void handle_interrupt();
   
   bool load_state();
   bool save_state();   
   //--------------------------------------------
   int button_pressed[8];
   
   int MBC1memorymodel;
   int RAMenable;
   int rom_bank;
   int ram_bank;
   int wram_bank; 
   int vram_bank;
   unsigned short MBChi;
   unsigned short MBClo;
   int RTCIO;   
   
   int skip_frame;
   int off_counter;
   int serialbits;
   int IWait;

   int STAT_wait; 
   int draw_wait;   
   int frameskip_counter;
   int mode3_cyc;
   int wy_set;
   int IME;
   int CPUHalt;
   int gb_speed;
   int int_line;
   int timer_freq;
   
   int gbc_mode;

   int cycles_DIV;
   int cycles_timer;
   int cycles_LCD;
   int CYCLES_SERIAL;
   int cycles_serial;  
   int cycles_sound;
   int EI_count; 
   
   int timeron;
   int LCDon;
   
   int hdma_source;
   int hdma_destination;
   int hdma_bytes;
   int hdma_on;
   int gdma_rest;
   
   int tile_map;
   int win_tile_map;
   int tile_pattern;
   byte BGP[4];
   byte OBP0[4];
   byte OBP1[4];
   int sprite_number;
   unsigned short GBC_BGP[32];
   unsigned short GBC_OBP[32];
   int windowline;
   
   byte* memory;
   byte *cartRAM;
   byte *VRAM;
   byte *WRAM;
   byte* mem_map[0x10];
   byte* cartridge;
   GBrom* rom;
     
   void* gfx_buffer1;
   void* gfx_buffer2;
   void* gfx_buffer_old;
   void* gfx_buffer;
   
   bool init_gfx();
   static int gfx_bit_count;   

   void set_LCD_mode(int mode) { memory[0xFF41] = (memory[0xFF41]&0xFC)|mode; }      
   void do_hdma(); 
   void set_int(int number) // sets interrupt 
   {
      if(number == 2) // allow only once per line
      {
         if(int_line == memory[0xFF44])
            return;
         else
            int_line = memory[0xFF44];
      }
   
      memory[0xFF0F] |= number;   
   }
   
   void compareLYCtoLY()
   {
      if(memory[0xFF45] == memory[0xFF44])
         memory[0xFF41] |= 0x04;
      else
         memory[0xFF41] &= 0xFB; 
   }
   
   int mainloop();
};

extern gb_system *GB;
extern gb_system *GB1;
extern gb_system *GB2;

#endif
