/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include <stdio.h>
#include <dplay.h>
#include "main.h"
#include "mainloop.h"
#include "cpu.h"
#include "debug.h"
#include "directplay.h"
#include "directdraw.h"
#include "mem.h"

IDirectPlay4* DP;
GUID AppGUID = {
                0xD351D351,
                0x99AB,
                0x1983,
                0x17,
                0xB0,
                0x18,
                0x25,
                0x77,
                0x41,
                0xF1,
                0x00};

GUID sessionGUID;
//char session_name[100];
bool use_protocol;
bool host_player;
char player_name[100];
int game_found = 0;
int net_initted = 0;
DPID LocalPlayerDPID;
DPID LinkedPlayerDPID;
#define MAX_CONNECTIONS 8
void* connections[MAX_CONNECTIONS];

BOOL FAR PASCAL DPConnect_EnumConnectionsCallback(LPCGUID pguidSP,VOID* pConnection,DWORD dwConnectionSize,LPCDPNAME pName,DWORD dwFlags,VOID* pvContext )
{
    static int connection_index;
    HRESULT       hr;
    IDirectPlay4* pDP = NULL;
    VOID*         pConnectionBuffer = NULL;
    HWND          hWndListBox = (HWND)pvContext;
    LRESULT       iIndex;
 
    // Create a IDirectPlay object.
    if(FAILED(hr = CoCreateInstance(CLSID_DirectPlay,NULL,CLSCTX_ALL,IID_IDirectPlay4A,(VOID**)&pDP)))
       return FALSE; // Error, stop enumerating
 
    // Test the if the connection is available 
    // by attempting to initialize.
    // the connection
    if(FAILED(hr = pDP->InitializeConnection(pConnection,0)))
    {
        SafeRelease(pDP);
        return TRUE; // Unavailable connection, keep enumerating
    }
 
    // Don't need the IDirectPlay interface any more, so release it.
    SafeRelease(pDP);
 
    // Found a good connection, so put it in the listbox.
    iIndex = SendMessage( hWndListBox, LB_ADDSTRING, 0,(LPARAM)pName->lpszShortNameA );
    if(iIndex == CB_ERR)
        return FALSE; // Error, stop enumerating
 
    //pConnectionBuffer = new BYTE[dwConnectionSize];
    //if(pConnectionBuffer == NULL)
    //    return FALSE; // Error, stop enumerating
 
    // Store pointer to GUID in listbox.
    //memcpy(pConnectionBuffer,pConnection,dwConnectionSize );
    //SendMessage(hWndListBox,LB_SETITEMDATA,iIndex,(LPARAM)pConnectionBuffer);
    connections[connection_index] = malloc(dwConnectionSize);
    memcpy(connections[connection_index],pConnection,dwConnectionSize);
    if(++connection_index>=MAX_CONNECTIONS)
       return FALSE;
    
    return TRUE; // Keep enumerating
}

bool init_connection(int index)
{
   HRESULT hr;
   if(FAILED(hr = DP->InitializeConnection(connections[index],0)))
   {
      if(hr==DPERR_ALREADYINITIALIZED) 
         return true;
      debug_print("Connection init failed.");
      return false; 
   }   
   return true;
}

bool Create_Session()
{
   DPSESSIONDESC2 dpsd;
   HRESULT hr;
   
   ZeroMemory(&dpsd,sizeof(dpsd));
   dpsd.dwSize           = sizeof(dpsd);
   dpsd.guidApplication  = AppGUID;
   dpsd.lpszSessionNameA = "GEST netplay game";//session_name;
   dpsd.dwMaxPlayers     = 2;
   // The DPSESSION_KEEPALIVE flag keeps the session alive 
   //  if players abnormally exit.
   dpsd.dwFlags          = DPSESSION_KEEPALIVE | DPSESSION_MIGRATEHOST;
   //if(use_protocol)
      dpsd.dwFlags |= DPSESSION_DIRECTPLAYPROTOCOL;
 
   // Create a new session.
   host_player = true;
   if(FAILED(hr = DP->Open(&dpsd, DPOPEN_CREATE)))
   {
      if(hr == DPERR_USERCANCEL)
         return false;
      
      debug_print("Create game failed!");
      return false;
   }
      
   DPNAME dpname;
   ZeroMemory( &dpname, sizeof(DPNAME) );
   dpname.dwSize         = sizeof(DPNAME);
   dpname.lpszShortNameA = player_name;
 
   if(FAILED(DP->CreatePlayer(&LocalPlayerDPID,&dpname,NULL,NULL,0,0)))
   {
      debug_print("Create player failed!");
      return false;
   }
      
   return true;
}

BOOL FAR PASCAL DPConnect_EnumSessionsCallback(LPCDPSESSIONDESC2 pdpsd,DWORD* pdwTimeout,DWORD dwFlags,VOID* pvContext)
{
    //DPSessionInfo* pDPSINew = NULL;
 
    if(dwFlags & DPESC_TIMEDOUT)
    {
       game_found = 0;
       return FALSE; // Timed out, so stop the enumeration.
    }
 
    // Found a good session, save it.
    //pDPSINew = new DPSessionInfo;
    //if(NULL == pDPSINew)
    //   return FALSE;
 
    //ZeroMemory(pDPSINew, sizeof(DPSessionInfo));
 
    // Copy the information into pDPSINew.
    sessionGUID = pdpsd->guidInstance;
    //sprintf(pDPSINew->szSession, "%s (%d/%d)", pdpsd->lpszSessionNameA,pdpsd->dwCurrentPlayers,pdpsd->dwMaxPlayers );
 
    // Add pDPSINew to the circular linked list, g_pDPSIFirst.
    //pDPSINew->pDPSINext = g_DPSIHead.pDPSINext;
    //g_DPSIHead.pDPSINext = pDPSINew;
    
    game_found = 1;
    return FALSE;
}

bool Join_Session()
{
   DPSESSIONDESC2 dpsd;
   HRESULT hr;
   // Enum sessions and let DirectPlay decide the timeout.
   ZeroMemory(&dpsd,sizeof(dpsd));
   dpsd.dwSize          = sizeof(dpsd);
   dpsd.guidApplication = AppGUID;
 
   // Enumerate all the active DirectPlay games 
   //  on the selected connection.
   hr = DP->EnumSessions(&dpsd,0,DPConnect_EnumSessionsCallback,NULL,DPENUMSESSIONS_ALL|DPENUMSESSIONS_ASYNC);  
   if(hr == DPERR_USERCANCEL)
      return false;
      
   if(!game_found) 
      return true;
      
   ZeroMemory(&dpsd,sizeof(dpsd));
   dpsd.dwSize          = sizeof(dpsd);
   dpsd.guidInstance    = sessionGUID;
   dpsd.guidApplication = AppGUID;
 
   // Join the session
   host_player = false;
   if(FAILED(hr = DP->Open(&dpsd,DPOPEN_JOIN)))
   {
      if(hr==DPERR_NONEWPLAYERS)
         debug_print("Game already has two players");
      else 
         debug_print("Join game failed.");
      return false;
   }
      
   DPNAME dpname;
   ZeroMemory(&dpname,sizeof(DPNAME));
   dpname.dwSize         = sizeof(DPNAME);
   dpname.lpszShortNameA = player_name;
 
   if(FAILED(DP->CreatePlayer(&LocalPlayerDPID,&dpname,NULL,NULL,0,0)))
   {
      debug_print("Create player failed!");
      return false;
   }
            
   return false;
}

bool Init_DP(HWND hwndDlg)
{
   HWND hwndlist = GetDlgItem(hwndDlg, ID_NET_SERVICE);

   if(net_initted)
   { 
      if(FAILED(DP->EnumConnections(&AppGUID,DPConnect_EnumConnectionsCallback,hwndlist,0)))
      {
         SafeRelease(DP);
         return false;
      }
      return true;
   }
   
   if(FAILED(CoInitialize(NULL)))
      return false;
  
   if(FAILED(CoCreateInstance(CLSID_DirectPlay,NULL,CLSCTX_ALL,IID_IDirectPlay4A,(VOID**)&DP)))
      return false;
  
   // Enumerate all DirectPlay connections and store them 
   // in the listbox
   if(FAILED(DP->EnumConnections(&AppGUID,DPConnect_EnumConnectionsCallback,hwndlist,0)))
   {
      SafeRelease(DP);
      return false;
   }
   
   net_initted = 1;
   return true;
}

void Kill_DP()
{   
   for(int i=0;i<MAX_CONNECTIONS;i++)
   {
      if(connections[i] != NULL)
      {
         free(connections[i]);
         connections[i] = NULL;
      }
   }
   SafeRelease(DP);
   
   CoUninitialize();   
}

int data_sent = 0;
int link_on = 0;
int link_gb_internal = 0;
byte link_gb_link[2] = {0,0};
byte link_gb_last[3] = {0,0,0};

byte link_gb_clock[3] = {0,0,0};

void send_gb_link_state(int this_link_on,int internal_clock)
{
   link_gb_link[0] = this_link_on;
   link_gb_link[1] = internal_clock; 

   DP->Send(LocalPlayerDPID,DPID_ALLPLAYERS,0,link_gb_link,2);

   //DP->SendEx(LocalPlayerDPID,DPID_ALLPLAYERS,DPSEND_ASYNC|DPSEND_NOSENDCOMPLETEMSG,link_gb_link,2,0,0,NULL,NULL);   
}

/*
void send_gb_clock_state()
{
   link_gb_clock[0] = cycles_serial;
   link_gb_clock[1] = serialbits;
      
   DP->Send(LocalPlayerDPID,DPID_ALLPLAYERS,0,link_gb_clock,3);
}*/

void link_send_data(byte data)
{   
   DP->Send(LocalPlayerDPID,DPID_ALLPLAYERS,0,&data,1);
   //DP->SendEx(LocalPlayerDPID,DPID_ALLPLAYERS,DPSEND_ASYNC|DPSEND_NOSENDCOMPLETEMSG,&data,1,0,0,NULL,NULL);
}

byte data_r;
int data_rr;

void check_link_state(int this_link_on,int this_is_internal)
{
   if(!net_initted)
   {
      link_on = 0;
      data_sent = 1;
      return;
   }
   DPID to;
   HRESULT hr;
   DWORD size = 0;
   link_on = 0;
   int link_state_found = 0;
   data_rr = 0;

   hr = DP->Receive(&LinkedPlayerDPID,&to,DPRECEIVE_PEEK,NULL,&size);
   while(hr != DPERR_NOMESSAGES)
   {
      if(LinkedPlayerDPID==DPID_SYSMSG)
      {
         byte* sysdata = (byte*)malloc(size);
         DP->Receive(&LinkedPlayerDPID,&to,0,sysdata,&size);
         DPMSG_GENERIC* sys_message = (DPMSG_GENERIC*)sysdata;
         switch(sys_message->dwType)
         {
         case DPSYS_SESSIONLOST:
            sprintf(dx_message,"%s.","Disconnected: Session lost!");
            message_time = 70;
         break;
         case DPSYS_HOST:
            sprintf(dx_message,"%s.","Host lost. You are now the host.");
            message_time = 70;            
         break;
         }
         free(sysdata);

      } else if(size == 2)
      {
         DP->Receive(&LinkedPlayerDPID,&to,0,link_gb_link,&size);
                           
         if(link_gb_link[0] && this_link_on && (this_is_internal^link_gb_link[1]))
            link_on = 1;
         
         memcpy(link_gb_last,link_gb_link,2);
         link_gb_last[2] = 1; // we have a linked computer
         
         link_state_found = 1;

         link_gb_internal = link_gb_link[1];
      } else
      {
         byte* sysdata = (byte*)malloc(size);
         DP->Receive(&LinkedPlayerDPID,&to,0,sysdata,&size);
         if(size==1 && this_is_internal)
         {
            data_r = *sysdata;
            data_rr = 1;
         }
         free(sysdata); 
      }
      
      hr = DP->Receive(&LinkedPlayerDPID,&to,DPRECEIVE_PEEK,NULL,&size);
   }
   
   if(!link_state_found && link_gb_last[2] == 1) 
   {
            
      // Other computer hasn't sended new state and 
      // we are sure that we are linked, so use last state
      if(link_gb_last[0] && this_link_on && (this_is_internal^link_gb_last[1]))
         link_on = 1;
   }// else if(link_state_found)
    //  clear_messages();
}

void clear_messages()
{
   DPID to;
   DPID from;
   DWORD size = 500;
   byte* data = (byte*)malloc(size);
   while(1)
   {
      if(DP->Receive(&from,&to,0,data,&size) == DPERR_NOMESSAGES)
         break;
      size = 500;
   }
   free(data);
   data_sent = 0;
   link_gb_last[2] = 0;
}

int syncing = 0;
byte received_data = 0;

unsigned short link_receive_data(int is_external,byte *rdata)
{
   DPID to;
   DWORD size = 1;
   HRESULT hr;

   if(FAILED(hr = DP->Receive(&LinkedPlayerDPID,&to,0,&received_data,&size)))
   {
      switch(hr)
      {
      case DPERR_BUFFERTOOSMALL:
         if(LinkedPlayerDPID==DPID_SYSMSG) // system messages
         {
            byte* sysdata = (byte*)malloc(size);
            DP->Receive(&LinkedPlayerDPID,&to,0,sysdata,&size);
            DPMSG_GENERIC* sys_message = (DPMSG_GENERIC*)sysdata;
            switch(sys_message->dwType)
            {
            case DPSYS_SESSIONLOST:
               sprintf(dx_message,"%s.","Disconnected: Session lost!");
               message_time = 70;
            break;
            case DPSYS_HOST:
               sprintf(dx_message,"%s.","Host lost. You are now the host.");
               message_time = 70;    
            break;
            }
            free(sysdata);
            return link_receive_data(is_external,rdata);
         } else
         if(size == 2)
         {
            DP->Receive(&LinkedPlayerDPID,&to,0,link_gb_link,&size);

            //if(link_gb_link[0] && (memory[0xFF02]&0x80) && ((memory[0xFF02]&1)^(link_gb_link[1])))
            //   link_on = 1;
            //link_gb_internal = link_gb_link[1];

            return link_receive_data(is_external,rdata);
         }
      case DPERR_NOMESSAGES:
         if(is_external) // external clock 
         {
            if(rdata==NULL) 
               return 0xffff; // wait for data
            else     
               return *rdata;
         }
         else 
         {        
            if(rdata!=NULL) 
            {
               return *rdata; 
            }
            else 
            {
               if(link_on)
               {
                  //if(data_rr)
                  //   return data_r;
                  syncing = 1;
                  //return memory[0xFF01];
                  //Sleep(5);
                  //return link_receive_data(0,NULL);
                  return 0xffff;
               } else 
                  return 0xff;
            }
         }
      break;
      default:
         return 0xff;
      }
   } else
   {      
      return link_receive_data(is_external,&received_data);
   }
     
   return received_data;
}
