/*
GEST - Gameboy emulator
Copyright (C) 2003-2004 TM

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#define WIN32_LEAN_AND_MEAN

#include <windows.h>
#include <shellapi.h>
#include <commdlg.h>
#include <mmsystem.h>
#include <stdio.h>
#include <fmod.h>

#include <math.h>
#include <time.h>

#define DIRECTINPUT_VERSION 0x0700
#include <dinput.h>

#include "strings.h"
#include "cheats.h"
#include "config.h"
#include "dialogs.h"

#include "main.h"
#include "debug.h"
#include "types.h"
#include "cpu.h"
#include "directdraw.h"
#include "directinput.h"
#include <dplay.h>
#include "directplay.h"
#include "mainloop.h"
#include "sound.h"
#include "gfx.h"
#include "SGB.h"
#include "devices.h"
#include "mem.h"

#include "GB.h"
gb_system* GB;
gb_system* GB1;
gb_system* GB2;

char* prg_version = "0.85";

// Windows stuff ----------------------------------------
HWND hwnd = NULL;           
HWND hwndCtrl = NULL;    
HMENU menu;
HINSTANCE hinst;

LRESULT CALLBACK WindowProcedure (HWND, UINT, WPARAM, LPARAM);

char szClassName[] = "GESTclass";
char emu_title[] = "GEST";
char title_text[ROM_FILENAME_SIZE + 8];

int timer_id = 0;
int sizen_w,sizen_h; // Size needed for menu,borders,etc.

// Options ----------------------------------------------
int mix_frames = 0;
int speed_up_option = 0;
int color_option = COL_BW;
extern int frame_skip;
int sgb_border_option = 0;
int system_sound_on = 1;
int win_size = 2;
bool paused = false;
bool menupause = false;

int auto_frameskip = 0;

// Controls ------------------------------------------
int control_pressed = 0; // control key pressed 
int multi_key_config[4][17] =
{
 {
 DIK_Z,DIK_X,DIK_RETURN,DIK_BACK,DIK_A,DIK_S,DIK_Q,DIK_W,
 DIK_SPACE,DIK_UP,DIK_DOWN,DIK_LEFT,DIK_RIGHT,
 DIK_NUMPAD4,DIK_NUMPAD6,DIK_NUMPAD8,DIK_NUMPAD2
 } 
 // others are none by default
};

int current_controller = 0; // currently changing which controller? 

int *key_config =  multi_key_config[0];

int speedup = 0;
int use_joystick_input = -1;

extern int time_to_switch;

#ifdef ALLOW_DEBUG
bool debug = false;
#endif

#include "licnames.h"
int ramsize[9] = { 0, 2, 8, 32, 128, 64,64,64,8 }; // KBytes
int romsize(int number)
{
   switch(number)
   {
      case 0: return 32; 
      case 1: return 64; 
      case 2: return 128;  
      case 3: return 256; 
      case 4: return 512;   
      case 5: return 1024;  
      case 6: return 2048; 
      case 7: return 4096; 
      case 8: return 8192;
      case 0x52: return 1126; 
      case 0x53: return 1228; 
      case 0x54: return 1536; 
   }
}

void PrintROMinfo()
{
   char text[500];
   //sprintf(text, "Rom name: \t%s\nGBC feature: \t%s\nNew Licensee: \t%X\nSGB feature: \t%s\nCardridge Type: \t%X\nROM Size: \t%d KBytes\nRAM Size: \t%d KBytes\nCountry: \t\t%X ,%s\nLicensee: \t%X - %s\nVersion: \t\t%X\nComplement check: %X - %s\nChecksum: \t%X", rom->name, rom->CGB==1?"yes":rom->CGB==2?"GBC only":"no", rom->newlic, rom->SGB?"yes":"no", rom->carttype,romsize(rom->ROMsize),ramsize[rom->RAMsize], rom->destcode, rom->destcode==0?"Japan":"non-Japan",rom->lic,lic_names[rom->lic],rom->version,rom->complement,rom->complementok?"(ok)":"(wrong)",rom->checksum);
   sprintf(text, str_table[ROM_INFO], GB1->rom->name, GB1->rom->CGB==1?str_table[STR_YES]:GB1->rom->CGB==2?str_table[GBC_ONLY]:str_table[STR_NO], GB1->rom->newlic, GB1->rom->SGB?str_table[STR_YES]:str_table[STR_NO], GB1->rom->carttype,romsize(GB1->rom->ROMsize),ramsize[GB1->rom->RAMsize], GB1->rom->destcode, GB1->rom->destcode==0?str_table[STR_JAPAN]:str_table[STR_NON_JAPAN],GB1->rom->lic,lic_names[GB1->rom->lic],GB1->rom->version,GB1->rom->complement,GB1->rom->complementok?str_table[CHECK_OK]:str_table[CHECK_WRONG],GB1->rom->checksum);
   debug_print(text);
}

void cleanup()
{
   if(GB1 != NULL) 
   {
      delete GB1;
      GB1 = NULL;
   }
   
   if(GB2 != NULL) 
   {
      delete GB2;
      GB2 = NULL;
   }   
                  
   sgb_end();
   
   Kill_DD(); 
   
   Kill_DI();  
   
   Kill_DP(); 
   
   FSOUND_Close();
}

int WINAPI WinMain(HINSTANCE hThisInstance,HINSTANCE hPrevInstance,LPSTR lpszArgument,int nFunsterStil)
{
   WNDCLASSEX wincl;        
   MSG msg;    
   
   /* The Window structure */
   hinst = hThisInstance;
   wincl.hInstance = hThisInstance;
   wincl.lpszClassName = szClassName;
   wincl.lpfnWndProc = WindowProcedure;      
   wincl.style = CS_DBLCLKS;                 /* Catch double-clicks */
   wincl.cbSize = sizeof(WNDCLASSEX);

   wincl.hIcon = LoadIcon(hThisInstance, MAKEINTRESOURCE(ID_ICON));
   wincl.hIconSm = LoadIcon(hThisInstance, MAKEINTRESOURCE(ID_ICON));
   wincl.hCursor = LoadCursor(NULL, IDC_ARROW);
   wincl.lpszMenuName = NULL;                 /* No menu */
   wincl.cbClsExtra = 0;                      /* No extra bytes after the window class */
   wincl.cbWndExtra = 0;                      /* structure or the window instance */
   wincl.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
   
   if(!RegisterClassEx(&wincl))
      return 0;
   
   menu = LoadMenu(hThisInstance, MAKEINTRESOURCE(ID_MENU));

   hwnd = CreateWindowEx(0,szClassName,emu_title,WS_SIZEBOX|WS_OVERLAPPEDWINDOW,150,150,2*160,2*144,HWND_DESKTOP,menu,hThisInstance,NULL);
   
   RECT adjrect;
   GetClientRect(hwnd,&adjrect);
   
   sizen_w = 2*160-(adjrect.right-adjrect.left);
   sizen_h = 2*144-(adjrect.bottom-adjrect.top);
    
   GB1 = new gb_system;
   if(!GB1 || !GB1->init() || !sgb_init())
   { 
      debug_print(str_table[ERROR_MEMORY]); 
      PostMessage(hwnd, WM_QUIT, 0, 0L);  
   }
   GB = GB1;
   
   if(!Init_DD()) 
   { 
      debug_print(str_table[ERROR_DDRAW]); 
      PostMessage(hwnd, WM_QUIT, 0, 0L);  
   }
        
   double time_elapsed=0.0;
   
   LARGE_INTEGER cur_time;
   DWORD time_count=16;
   LARGE_INTEGER perf_cnt;
   LARGE_INTEGER next_time;
   LARGE_INTEGER last_time;
   int perf_flag = 0;
     
   if(QueryPerformanceFrequency(&perf_cnt))
   {
      perf_flag = 1;
      time_count = (DWORD)(perf_cnt.QuadPart/59.7275);
      QueryPerformanceCounter(&next_time);
   } else 
   {
      next_time.QuadPart = timeGetTime();
      time_count = 33;
   }

   //get program directory
   memset(program_directory,0,ROM_PATH_SIZE);
   char buffer[ROM_PATH_SIZE];
   GetModuleFileName(NULL,buffer,ROM_PATH_SIZE);   
   char* ptr = strrchr(buffer,'\\'); // without the program name
   strncpy(program_directory,buffer,strlen(buffer)-strlen(ptr));

   if(!read_config_file())
      debug_print(str_table[ERROR_CFG_FILE_READ]);

   ShowWindow(hwnd, SW_SHOW);//nFunsterStil);     

   if(!Init_DI())
   {
      debug_print(str_table[ERROR_DINPUT]);
      PostMessage(hwnd, WM_QUIT, 0, 0L);        
   }

   FSOUND_SetHWND(hwnd);
   if(FSOUND_Init(44100,2,0) == FALSE)
   {
       debug_print(str_table[ERROR_FMOD]);
       system_sound_on = -1;     
   } else
      FSbuffer = FSOUND_Sample_Alloc(FSOUND_FREE,sound_buffer_total_len>>2,FSOUND_LOOP_NORMAL|FSOUND_16BITS|FSOUND_STEREO|FSOUND_SIGNED,44100,200,FSOUND_STEREOPAN,1);
   
   if(system_sound_on)
      channel_n = FSOUND_PlaySound(FSOUND_FREE,FSbuffer);

   DragAcceptFiles(hwnd,TRUE);
   
   double time_scale;
   
   if(strlen(lpszArgument)>4)
   {
      char* ptr = strchr(lpszArgument,'"');
      if(ptr != NULL) // does the filename contain " " ?
      {
         ptr += 1;
         char load_filename[600];
         memset(load_filename,0,600);
         strncpy(load_filename,ptr,strlen(ptr)-1);
         GB1->load_rom(load_filename); 
      } else
         GB1->load_rom(lpszArgument); 
      
      if(GB1->romloaded)
      {
         reset(1);
              
         if(sgb_mode)
         {
            RECT winRect;
            GetWindowRect(hwnd,&winRect);      
            MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
         } else
         {
            RECT winRect;
            GetWindowRect(hwnd,&winRect);      
            MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);                     
         }
                     
         if(!read_save(1))
            debug_print(str_table[ERROR_SAVE_FILE_READ]);
      
         sprintf(title_text,"%s - %s",emu_title,GB1->rom_filename);
         SetWindowText(hwnd,title_text);
      }
   }

   for(;;)
   {
      if(PeekMessage(&msg,NULL,0,0,PM_REMOVE)) 
      {
         if(msg.message == WM_QUIT)
            break;
         TranslateMessage(&msg);
         DispatchMessage(&msg);
      } else 
      {        
         #ifdef ALLOW_DEBUG                
         if(debug && !paused && romloaded[1])
         {
            frames = 1;
            mainloop();
            frames = 0;

            draw_debug_screen();
            continue;
         }
         #endif
         
         if(!emulating || !GB1->romloaded || paused)
         {
            Sleep(50);
            continue;
         }
       
         if(perf_flag)
            QueryPerformanceCounter(&cur_time);
         else
            cur_time.QuadPart = timeGetTime();
         
         if((system_sound_on > 0) || speedup || cur_time.QuadPart >= next_time.QuadPart)
         {
             // runs until frame
             do
             {
                GB->mainloop();   
                if(time_to_switch == 1)
                {
                   switch_gb();
                   time_to_switch = 0;
                }
             } while(frames==0);

             if(multiple_gb) switch_gb();

             next_time.QuadPart = cur_time.QuadPart+(multiple_gb?time_count/2:time_count);
             
             if(auto_frameskip)
             {
                QueryPerformanceCounter(&cur_time);
                if(cur_time.QuadPart > next_time.QuadPart)
                {
                   if(frame_skip < 9)
                   {
                      ++frame_skip;
                   }
                }
                else if(frame_skip > 0)
                {
                   --frame_skip;                                                 
                }
             } 
         }

         check_system_keys();

         frames = 0;
      }
   }

   return msg.wParam;
}

LRESULT CALLBACK WindowProcedure(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
    switch (message)           
    {
    case WM_COMMAND:
           switch( wParam )
           {
             case IDM_FILEOPEN:
             {                  
                  bool romwasloaded = GB1->romloaded;
                  
                  if(GB1->romloaded && !write_save(1))
                     debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                  
                  if(!DoFileOpen(hwnd,1)) 
                     debug_print(str_table[ERROR_ROM_LOAD]);
                                    
                  if(GB1->romloaded)
                  {            
                     if(multiple_gb) init_gb1();      
                     reset(1);
                     
                     if(!multiple_gb)
                     {
                        if(sgb_mode)
                        {
                           RECT winRect;
                           GetWindowRect(hwnd,&winRect);      
                           MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
                        } else
                        {
                           RECT winRect;
                           GetWindowRect(hwnd,&winRect);      
                           MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);                     
                        }
                     }

                     if(!read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);

                     if(multiple_gb)
                        sprintf(title_text,"%s - %s --- %s",emu_title,GB1->rom_filename,GB2->rom_filename);
                     else
                        sprintf(title_text,"%s - %s",emu_title,GB1->rom_filename);
                     SetWindowText(hwnd,title_text);
                     if(!paused)
                        FSOUND_SetMute(FSOUND_ALL,FALSE);   
                  } else 
                  if(romwasloaded && GB1->cartridge != NULL) 
                  {
                     GB1->romloaded = true;
                     if(!paused)
                        FSOUND_SetMute(FSOUND_ALL,FALSE);                      
                  }
             }               
             break;

             case IDM_FILEOPEN2:
             {
                  init_gb2();
                  
                  bool romwasloaded = GB2->romloaded;
                  
                  if(GB2->romloaded && !write_save(2))
                     debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                  
                  if(!DoFileOpen(hwnd,2)) 
                     debug_print(str_table[ERROR_ROM_LOAD]);
                  
                  if(GB2->romloaded)
                  {        
                     init_gb2();             
                     reset(2);
 
                     if(!read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);
                     
                     sprintf(title_text,"%s - %s --- %s",emu_title,GB1->rom_filename,GB2->rom_filename);
                     SetWindowText(hwnd,title_text);
                     if(!paused)
                        FSOUND_SetMute(FSOUND_ALL,FALSE);   
                  } else 
                  if(romwasloaded && GB2->cartridge != NULL) 
                  {
                     GB2->romloaded = true;
                     if(!paused)
                        FSOUND_SetMute(FSOUND_ALL,FALSE);                      
                  }
             }            
             break;
                  
             case IDM_FILECLOSE:
                  FSOUND_SetMute(FSOUND_ALL,TRUE);
                  SetWindowText(hwnd,emu_title);
                  if(GB1->romloaded)
                  {
                     if(!write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                  }
                  if(GB1->cartridge)
                  {
                     delete [] GB1->cartridge;
                     GB1->cartridge = NULL;
                  }
                  GB1->romloaded = false;
                  memset(GB1->gfx_buffer,0,sizeof(unsigned short)*160*144);
                  memset(GB1->gfx_buffer_old,0,sizeof(unsigned short)*160*144);  
                  draw_screen();                
             break;

             case IDM_FILECLOSE2:
                  if(!GB2) init_gb2();
                  if(current_gb==2)
                     switch_gb();
                  if(!GB1->romloaded) 
                  {
                     FSOUND_SetMute(FSOUND_ALL,TRUE);
                     SetWindowText(hwnd,emu_title);
                  }
                  RECT winRect;
                  GetWindowRect(hwnd,&winRect);      
                  MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);

                  if(GB2->romloaded)
                  {
                     if(!write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                  }
                  if(GB2->cartridge)
                  {
                     delete [] GB2->cartridge;
                     GB2->cartridge = NULL;
                  }
                  GB2->romloaded = false;
                  multiple_gb = 0;    
                  
                  POINT pt;
                  GetClientRect(hwnd,&target_blt_rect);
                  pt.x=pt.y=0;
                  ClientToScreen(hwnd,&pt);
                  OffsetRect(&target_blt_rect,pt.x,pt.y);       
             break;             
             
             case IDM_ROMINFO:
                  if(GB1->romloaded)
                     PrintROMinfo();
             break;
             
             case IDM_NETPLAY:
                DialogBox(hinst, MAKEINTRESOURCE(ID_NETPLAY_DIALOG), hwnd, (DLGPROC)NetplayProc);
             break;     
                          
             case IDM_FOLDERS:
                DialogBox(hinst, MAKEINTRESOURCE(ID_F_DIALOG), hwnd, (DLGPROC)FolderProc);
             break;     
             
             case IDM_FILEZIPNAME:
                if(zip_name_option==USE_ZIPFILE_NAME)
                {
                   CheckMenuItem(menu,IDM_FILEZIPNAME,MF_UNCHECKED);
                   zip_name_option=USE_ROMFILE_NAME;
                } else
                {
                   CheckMenuItem(menu,IDM_FILEZIPNAME,MF_CHECKED);
                   zip_name_option=USE_ZIPFILE_NAME;              
                } 
             break;              
             case IDM_CHEATS:
                DialogBox(hinst, MAKEINTRESOURCE(ID_CHT_DIALOG), hwnd, (DLGPROC)ChtProc);               
             break;
             case IDM_FILEEXIT:
                  SendMessage(hwnd, WM_CLOSE, 0, 0L);
             break;
             
             case IDM_CPUSOFTRESET:
                   soft_reset=1;
             break;
             
             case IDM_CPURESET:
                  if(GB1->romloaded && !write_save(1))
                     debug_print(str_table[ERROR_SAVE_FILE_WRITE]);

                  reset(1);

                  if(GB1->romloaded && !read_save(1))
                     debug_print(str_table[ERROR_SAVE_FILE_READ]);
             break;
             
             case IDM_CPURESET2:
                  if(!GB2) init_gb2();
                  
                  if(GB2->romloaded && !write_save(2))
                     debug_print(str_table[ERROR_SAVE_FILE_WRITE]);

                  reset(2);

                  if(GB2->romloaded && !read_save(2))
                     debug_print(str_table[ERROR_SAVE_FILE_READ]);
             break;
                          
             case IDM_CPUPAUSE:
                if(menupause)
                {
                   if(GB1->romloaded)
                      FSOUND_SetMute(FSOUND_ALL,FALSE);
                   CheckMenuItem(menu,IDM_CPUPAUSE,MF_UNCHECKED);
                }
                else
                {
                   FSOUND_SetMute(FSOUND_ALL,TRUE);
                   CheckMenuItem(menu,IDM_CPUPAUSE,MF_CHECKED); 
                }
                menupause=!menupause;
                paused=!paused;
             break;
             case IDM_CPUSTATELOAD:
                if(GB1->romloaded)
                   GB1->load_state();
             break;
             case IDM_CPUSTATESAVE:
                if(GB1->romloaded)
                   GB1->save_state();
             break;      
             case IDM_CPUSTATESLOT0:
                state_slot = 0;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break;       
             case IDM_CPUSTATESLOT1:
                state_slot = 1;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT2:
                state_slot = 2;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT3:
                state_slot = 3;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT4:
                state_slot = 4;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT5:
                state_slot = 5;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT6:
                state_slot = 6;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT7:
                state_slot = 7;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_CHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT8:
                state_slot = 8;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_CHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_UNCHECKED);                 
             break; 
             case IDM_CPUSTATESLOT9:
                state_slot = 9;
                CheckMenuItem(menu,IDM_CPUSTATESLOT0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_CPUSTATESLOT8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_CPUSTATESLOT9,MF_CHECKED);                 
             break;       

             case IDM_CPUAUTO2:
                  if(!GB2) init_gb2();
                  if(GB2->system_type != SYS_AUTO)
                  {
                     GB2->system_type = SYS_AUTO;                  
                     
                     if(GB2->romloaded && !write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(2);
                     if(GB2->romloaded && !read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);         
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO2,MF_CHECKED);
                  
                  CheckMenuItem(menu,IDM_CPUGB2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC2,MF_UNCHECKED);                                                                          
             break;                  
             case IDM_CPUGBA2:
                  if(!GB2) init_gb2();
                  if(GB2->system_type != SYS_GBA)
                  {
                     GB2->system_type = SYS_GBA;            
                           
                     if(GB2->romloaded && !write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(2);
                     if(GB2->romloaded && !read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);         
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC2,MF_UNCHECKED);                                     
                                    
                  CheckMenuItem(menu,IDM_CPUGBA2,MF_CHECKED);            
                  break;
             case IDM_CPUGB2:
                  if(!GB2) init_gb2();
                  if(GB2->system_type != SYS_GB)
                  {
                     GB2->system_type = SYS_GB;         
                              
                     if(GB2->romloaded && !write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(2);
                     if(GB2->romloaded && !read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                 
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC2,MF_UNCHECKED);                                     
                                    
                  CheckMenuItem(menu,IDM_CPUGB2,MF_CHECKED);            
                  break;                
             case IDM_CPUGBP2:
                  if(!GB2) init_gb2();
                  if(GB2->system_type != SYS_GBP)
                  {
                     GB2->system_type = SYS_GBP; 
                                      
                     if(GB2->romloaded && !write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(2);
                     if(GB2->romloaded && !read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                 
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC2,MF_UNCHECKED);                   

                  CheckMenuItem(menu,IDM_CPUGBP2,MF_CHECKED);            
             break;       
             case IDM_CPUGBC2:
                  if(!GB2) init_gb2();
                  if(GB2->system_type != SYS_GBC)
                  {
                     GB2->system_type = SYS_GBC;     
                                  
                     if(GB2->romloaded && !write_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(2);
                     if(GB2->romloaded && !read_save(2))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB2,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP2,MF_UNCHECKED);  
                  CheckMenuItem(menu,IDM_CPUGBA2,MF_UNCHECKED);
                                    
                  CheckMenuItem(menu,IDM_CPUGBC2,MF_CHECKED);            
             break;   

                                                                                                                                                       
             case IDM_CPUAUTO:
                  if(GB1->system_type != SYS_AUTO)
                  {
                     GB1->system_type = SYS_AUTO;                  
                     
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);         
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_CHECKED);
                  
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED);                                                 
                  break;                  
             case IDM_CPUGBA:
                  if(GB1->system_type != SYS_GBA)
                  {
                     GB1->system_type = SYS_GBA;            
                           
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);         
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);                                     
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED); 
                                    
                  CheckMenuItem(menu,IDM_CPUGBA,MF_CHECKED);            
                  break;
             case IDM_CPUGB:
                  if(GB1->system_type != SYS_GB)
                  {
                     GB1->system_type = SYS_GB;         
                              
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                 
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);                                     
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED); 
                                    
                  CheckMenuItem(menu,IDM_CPUGB,MF_CHECKED);            
                  break;                
             case IDM_CPUGBP:
                  if(GB1->system_type != SYS_GBP)
                  {
                     GB1->system_type = SYS_GBP; 
                                      
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                 
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);                   
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED); 
                                    
                  CheckMenuItem(menu,IDM_CPUGBP,MF_CHECKED);            
             break;       
             case IDM_CPUGBC:
                  if(GB1->system_type != SYS_GBC)
                  {
                     GB1->system_type = SYS_GBC;     
                                  
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);  
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED); 
                                    
                  CheckMenuItem(menu,IDM_CPUGBC,MF_CHECKED);            
             break;   
             case IDM_CPUSGB:
                  if(GB1->system_type != SYS_SGB)
                  {
                     GB1->system_type = SYS_SGB;    
                                   
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);             
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);  
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  
                  CheckMenuItem(menu,IDM_CPUSGB,MF_CHECKED);     
                                            
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_UNCHECKED);                   
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);            
             break;  
             case IDM_CPUSGB2:
                  if(GB1->system_type != SYS_SGB2)
                  {
                     GB1->system_type = SYS_SGB2;    
                                   
                     if(GB1->romloaded && !write_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     reset(1);
                     if(GB1->romloaded && !read_save(1))
                        debug_print(str_table[ERROR_SAVE_FILE_READ]);                
                  }
                  CheckMenuItem(menu,IDM_CPUAUTO,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGB,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUGBP,MF_UNCHECKED);  
                  CheckMenuItem(menu,IDM_CPUGBA,MF_UNCHECKED);
                  CheckMenuItem(menu,IDM_CPUSGB,MF_UNCHECKED);  
                                               
                  CheckMenuItem(menu,IDM_CPUSGB2,MF_CHECKED); 
                                    
                  CheckMenuItem(menu,IDM_CPUGBC,MF_UNCHECKED);            
             break;                                             
             case IDM_CPUGBCSGB:
                if(sgb_border_option == SGB_BORDER)
                {
                   CheckMenuItem(menu,IDM_CPUGBCSGB,MF_UNCHECKED);
                   sgb_border_option = 0;
                } else
                {
                   CheckMenuItem(menu,IDM_CPUGBCSGBI,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_CPUGBCSGB,MF_CHECKED);
                   sgb_border_option = SGB_BORDER;                
                }            
             break;         
             case IDM_CPUGBCSGBI:
                if(sgb_border_option == SGB_INITB)
                {
                   CheckMenuItem(menu,IDM_CPUGBCSGBI,MF_UNCHECKED);
                   sgb_border_option = 0;
                } else
                {
                   CheckMenuItem(menu,IDM_CPUGBCSGB,MF_UNCHECKED);                
                   CheckMenuItem(menu,IDM_CPUGBCSGBI,MF_CHECKED);
                   sgb_border_option = SGB_INITB;                
                }                    
             break;      
             case IDM_CPUSPFS9:
                if(speed_up_option&SP_FS9)
                {
                   CheckMenuItem(menu,IDM_CPUSPFS9,MF_UNCHECKED);
                   speed_up_option &= ~SP_FS9;
                } else
                {
                   CheckMenuItem(menu,IDM_CPUSPFS9,MF_CHECKED);
                   speed_up_option |= SP_FS9;                
                }                
             break;    
             case IDM_CPUSPSNDOFF:
                if(speed_up_option&SP_SNDOFF)
                {
                   CheckMenuItem(menu,IDM_CPUSPSNDOFF,MF_UNCHECKED);
                   speed_up_option &= ~SP_SNDOFF;
                } else
                {
                   CheckMenuItem(menu,IDM_CPUSPSNDOFF,MF_CHECKED);
                   speed_up_option |= SP_SNDOFF;                
                }                
             break;       
             case IDM_CPUOPCODE:
                if(unk_opcode_option)
                {
                   CheckMenuItem(menu,IDM_CPUOPCODE,MF_UNCHECKED);
                   unk_opcode_option = 0;
                } else
                {
                   CheckMenuItem(menu,IDM_CPUOPCODE,MF_CHECKED);
                   unk_opcode_option = 1;               
                }                
             break;          
             case IDM_DEVICENONE:
                connected_device = DEVICE_NONE;
                reset_devices();

                CheckMenuItem(menu,IDM_DEVICENONE,MF_CHECKED);
                CheckMenuItem(menu,IDM_DEVICEBARCODE,MF_UNCHECKED);                           
             break; 
             case IDM_DEVICEBARCODE:
                connected_device = DEVICE_BARCODE;
                reset_devices();

                CheckMenuItem(menu,IDM_DEVICENONE,MF_UNCHECKED);
                CheckMenuItem(menu,IDM_DEVICEBARCODE,MF_CHECKED);                           
             break;                                                                                
             case IDM_OPTIONVIDEOSIZE1:
             {
                  RECT winRect;
                  GetWindowRect(hwnd,&winRect);
                  win_size = 1;
                  if(!sgb_mode && !(sgb_border_option==SGB_INITB && border_uploaded))
                     MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                  else
                     MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);              
                  if(multiple_gb)
                     MoveWindow(hwnd,winRect.left,winRect.top,2*160*win_size+sizen_w,144*win_size+sizen_h,TRUE);                     
             }
             break;

             case IDM_OPTIONVIDEOSIZE2:
             {
                  RECT winRect;
                  GetWindowRect(hwnd,&winRect);             
                  win_size = 2;     
                  if(!sgb_mode && !(sgb_border_option==SGB_INITB && border_uploaded))
                     MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                  else
                     MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
                  if(multiple_gb)
                     MoveWindow(hwnd,winRect.left,winRect.top,2*160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
             }
             break;

             case IDM_OPTIONVIDEOSIZE3:
             {
                  RECT winRect;
                  GetWindowRect(hwnd,&winRect);             
                  win_size = 3;              
                  if(!sgb_mode && !(sgb_border_option==SGB_INITB && border_uploaded))
                     MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                  else
                     MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
                  if(multiple_gb)
                     MoveWindow(hwnd,winRect.left,winRect.top,2*160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                     
             }
             break;

             case IDM_OPTIONVIDEOSIZE4:
             {
                  RECT winRect;
                  GetWindowRect(hwnd,&winRect);             
                  win_size = 4;         
                  if(!sgb_mode && !(sgb_border_option==SGB_INITB && border_uploaded))
                     MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                  else
                     MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
                  if(multiple_gb)
                     MoveWindow(hwnd,winRect.left,winRect.top,2*160*win_size+sizen_w,144*win_size+sizen_h,TRUE);
                     
             }
             break;
                  
             case IDM_OPTIONVIDEOBW: 
                color_option = COL_BW;
                CheckMenuItem(menu,IDM_OPTIONVIDEOBW,MF_CHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDBROWN,MF_UNCHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDGREEN,MF_UNCHECKED);	
                CheckMenuItem(menu,IDM_OPTIONVIDEOGRAY,MF_UNCHECKED);

                if(!GB->gbc_mode || !GB->rom->CGB)
                {
                   memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[0],sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[0],sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[0],sizeof(unsigned int)*4);
                }                   
             break;   
                  
             case IDM_OPTIONVIDEOLCDBROWN:  
                color_option = COL_LCDBROWN;             
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDBROWN,MF_CHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOBW,MF_UNCHECKED);	 
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDGREEN,MF_UNCHECKED);	
                CheckMenuItem(menu,IDM_OPTIONVIDEOGRAY,MF_UNCHECKED);                
                
                if(!GB->gbc_mode || !GB->rom->CGB)
                {
                   memcpy(GB->GBC_BGP,LCD_palette_brown,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP,LCD_palette_brown,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP+4,LCD_palette_brown,sizeof(unsigned int)*4);
                }                              
             break;    
             
             case IDM_OPTIONVIDEOLCDGREEN:  
                color_option = COL_LCDGREEN;             
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDGREEN,MF_CHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOBW,MF_UNCHECKED);	 
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDBROWN,MF_UNCHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOGRAY,MF_UNCHECKED);                	
                
                if(!GB->gbc_mode || !GB->rom->CGB)
                {
                   memcpy(GB->GBC_BGP,LCD_palette_green,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP,LCD_palette_green,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP+4,LCD_palette_green,sizeof(unsigned int)*4);
                }                              
             break;    

             case IDM_OPTIONVIDEOGRAY:  
                color_option = COL_GRAY;             
                CheckMenuItem(menu,IDM_OPTIONVIDEOGRAY,MF_CHECKED);
                
                CheckMenuItem(menu,IDM_OPTIONVIDEOBW,MF_UNCHECKED);	 
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDBROWN,MF_UNCHECKED);
                CheckMenuItem(menu,IDM_OPTIONVIDEOLCDGREEN,MF_UNCHECKED);                	
                
                if(!GB->gbc_mode || !GB->rom->CGB)
                {
                   memcpy(GB->GBC_BGP,GB_palette_gray,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP,GB_palette_gray,sizeof(unsigned int)*4);
                   memcpy(GB->GBC_OBP+4,GB_palette_gray,sizeof(unsigned int)*4);
                }                              
             break;    
                                       
             case IDM_VIDEOFILTERNONE:
                if(filter!=FILTER_NONE)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_UNCHECKED);              
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_UNCHECKED);                                                     
                                                     
                   filter = FILTER_NONE;
                   filter_width=1;
                   filter_height=1;
                   
                   change_filter();
                }
             break;
             case IDM_VIDEOFILTERSOFT2X:
                if(filter!=FILTER_SOFT2X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_UNCHECKED);                                  
                                                     
                   filter = FILTER_SOFT2X;
                   filter_width=2;
                   filter_height=2;
                   
                   change_filter();
                }
             break;     
             case IDM_VIDEOFILTERSCALE2X:
                if(filter!=FILTER_SCALE2X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_UNCHECKED);                    
                   CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_UNCHECKED);                                  
                   
                   filter = FILTER_SCALE2X;
                   filter_width=2;
                   filter_height=2;
                   
                   change_filter();
                }
             break;      
             case IDM_VIDEOFILTERSCALE3X:
                if(filter!=FILTER_SCALE3X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_UNCHECKED);                                  
                   
                   filter = FILTER_SCALE3X;
                   filter_width=3;
                   filter_height=3;
                   
                   change_filter();
                }
             break;              
             case IDM_VIDEOFILTERBLUR:
                if(filter!=FILTER_BLUR)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBLUR,MF_CHECKED);                
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERSOFT2X,MF_UNCHECKED); 
                                  
                   filter = FILTER_BLUR;
                   filter_width=2;
                   filter_height=2;
                   
                   change_filter();
                }
             break;                
             case IDM_VIDEOFILTERUSEMMX:
                if(use_mmx)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERUSEMMX,MF_UNCHECKED);
                   use_mmx = 0;
                } else
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERUSEMMX,MF_CHECKED);
                   use_mmx = 1;                
                }
             break;          

             case IDM_VIDEOFILTERBORDERNONE:
                if(border_filter!=FILTER_NONE)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_UNCHECKED);              
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_UNCHECKED);                                                     
                                                     
                   border_filter = FILTER_NONE;
                   border_filter_width=1;
                   border_filter_height=1;
                   
                   change_filter();
                }
             break;
             case IDM_VIDEOFILTERBORDERSOFT2X:
                if(border_filter!=FILTER_SOFT2X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_UNCHECKED);                                  
                                                     
                   border_filter = FILTER_SOFT2X;
                   border_filter_width=2;
                   border_filter_height=2;
                   
                   change_filter();
                }
             break;     
             case IDM_VIDEOFILTERBORDERSCALE2X:
                if(border_filter!=FILTER_SCALE2X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_CHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_UNCHECKED);                                  
                   
                   border_filter = FILTER_SCALE2X;
                   border_filter_width=2;
                   border_filter_height=2;
                   
                   change_filter();
                }
             break;     
             case IDM_VIDEOFILTERBORDERSCALE3X:
                if(border_filter!=FILTER_SCALE3X)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_CHECKED);                  
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_UNCHECKED);                                  
                   
                   border_filter = FILTER_SCALE3X;
                   border_filter_width = 3;
                   border_filter_height = 3;
                   
                   change_filter();
                }
             break;                 
             case IDM_VIDEOFILTERBORDERBLUR:
                if(border_filter!=FILTER_BLUR)
                {
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERBLUR,MF_CHECKED);                
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE2X,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSCALE3X,MF_UNCHECKED);                   
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERNONE,MF_UNCHECKED); 
                   CheckMenuItem(menu,IDM_VIDEOFILTERBORDERSOFT2X,MF_UNCHECKED); 
                                  
                   border_filter = FILTER_BLUR;
                   border_filter_width=2;
                   border_filter_height=2;
                   
                   change_filter();
                }
             break;                
                                    
             case IDM_OPTIONVIDEOMIX:
             if(mix_frames)
             {
                CheckMenuItem(menu,IDM_OPTIONVIDEOMIX,MF_UNCHECKED);
                
                mix_frames = 0;
                if(dx_bitcount == 16)
                   draw_screen = draw_screen16;
                else
                   draw_screen = draw_screen32;             	              
             } else
             {
                CheckMenuItem(menu,IDM_OPTIONVIDEOMIX,MF_CHECKED);
                
                mix_frames = 1;
                if(dx_bitcount == 16)
                   draw_screen = draw_screen_mix16;
                else
                   draw_screen = draw_screen_mix32;           	                           
             }
             break;
             
             case IDM_OPTIONVIDEOFSAUTO:
                if(auto_frameskip)
                {
                   auto_frameskip = 0;
                   CheckMenuItem(menu,IDM_OPTIONVIDEOFSAUTO,MF_UNCHECKED); 
                } else
                {
                   auto_frameskip = 1;
                   CheckMenuItem(menu,IDM_OPTIONVIDEOFSAUTO,MF_CHECKED);
                }               
             break;             
             case IDM_OPTIONVIDEOFS0:
                frame_skip = 0;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;
             case IDM_OPTIONVIDEOFS1:
                frame_skip = 1;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;             
             case IDM_OPTIONVIDEOFS2:
                frame_skip = 2;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;                             
             case IDM_OPTIONVIDEOFS3:
                frame_skip = 3;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;  
             case IDM_OPTIONVIDEOFS4:
                frame_skip = 4;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;              
             case IDM_OPTIONVIDEOFS5:
                frame_skip = 5;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;           
             case IDM_OPTIONVIDEOFS6:
                frame_skip = 6;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;            
             case IDM_OPTIONVIDEOFS7:
                frame_skip = 7;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_CHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;                     
             case IDM_OPTIONVIDEOFS8:
                frame_skip = 8;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_CHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_UNCHECKED);                 
             break;                  
             case IDM_OPTIONVIDEOFS9:
                frame_skip = 9;
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS0,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS3,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS4,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS5,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS6,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS7,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS8,MF_UNCHECKED);                                                                                                                                 
                CheckMenuItem(menu,IDM_OPTIONVIDEOFS9,MF_CHECKED);                 
             break;       
             case IDM_OPTIONVIDEOENBG:
                if(video_enable&VID_EN_BG)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENBG,MF_UNCHECKED);
                   video_enable &= ~VID_EN_BG;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENBG,MF_CHECKED);
                   video_enable |= VID_EN_BG;                
                }              
             break;   
             case IDM_OPTIONVIDEOENWIN:
                if(video_enable&VID_EN_WIN)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENWIN,MF_UNCHECKED);
                   video_enable &= ~VID_EN_WIN;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENWIN,MF_CHECKED);
                   video_enable |= VID_EN_WIN;                
                }              
             break;  
             case IDM_OPTIONVIDEOENSPRITE:
                if(video_enable&VID_EN_SPRITE)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENSPRITE,MF_UNCHECKED);
                   video_enable &= ~VID_EN_SPRITE;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENSPRITE,MF_CHECKED);
                   video_enable |= VID_EN_SPRITE;                
                }              
             break;  
             case IDM_OPTIONVIDEOSPRLIM:
                if(sprite_limit)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOSPRLIM,MF_UNCHECKED);
                   sprite_limit = false;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOSPRLIM,MF_CHECKED);
                   sprite_limit = true;                
                }              
             break;          
             case IDM_OPTIONVIDEOLCDOFF:
                if(LCDoff_option)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOLCDOFF,MF_UNCHECKED);
                   LCDoff_option = 0;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOLCDOFF,MF_CHECKED);
                   LCDoff_option = 1;                
                }              
             break;    
             case IDM_OPTIONVIDEOGBCCOLORS:
                if(gbc_real_colors)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOGBCCOLORS,MF_UNCHECKED);
                   gbc_real_colors = 0;
                   mix_gbc_colors();
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOGBCCOLORS,MF_CHECKED);                
                   gbc_real_colors = 1;
                   if(GB1->romloaded && GB1->gbc_mode)
                      mix_gbc_colors();
                }
             break;                      
             case IDM_OPTIONCONTROL1:
                current_controller = 0;
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_CTRL_DIALOG), hwnd, (DLGPROC) CtrlProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                 
                Init_DI_change(hwndCtrl);
             break; 
             case IDM_OPTIONCONTROL2:
                current_controller = 1;
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_CTRLMULTI_DIALOG), hwnd, (DLGPROC) MultiCtrlProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                 
                Init_DI_change(hwndCtrl);
             break; 
             case IDM_OPTIONCONTROL3:
                current_controller = 2;
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_CTRLMULTI_DIALOG), hwnd, (DLGPROC) MultiCtrlProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                 
                Init_DI_change(hwndCtrl);
             break; 
             case IDM_OPTIONCONTROL4:
                current_controller = 3;
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_CTRLMULTI_DIALOG), hwnd, (DLGPROC) MultiCtrlProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                 
                Init_DI_change(hwndCtrl);
             break;                                        
             case IDM_OPTIONSENSOR:
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_SENSOR_DIALOG), hwnd, (DLGPROC) SensorProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                
                Init_DI_change(hwndCtrl);
             break;        
             case IDM_OPTIONCONTROLJOYCH:
                hwndCtrl = CreateDialog(hinst, MAKEINTRESOURCE(ID_CTRLJOY_DIALOG), hwnd, (DLGPROC) JoyCtrlProc); 
                ShowWindow(hwndCtrl, SW_SHOW); 
                SetFocus(hwndCtrl);
                
                Init_DI_change_joy(hwndCtrl);
             break;                       
             case IDM_OPTIONCONTROLJOY1:
                if(use_joystick_input == 0)
                {
                   use_joystick_input = -1;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_UNCHECKED);
                } else
                {
                   use_joystick_input = 0;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_CHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_UNCHECKED);
                }
             break;    
             case IDM_OPTIONCONTROLJOY2:
                if(use_joystick_input == 1)
                {
                   use_joystick_input = -1;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_UNCHECKED);
                } else
                {
                   use_joystick_input = 1;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_CHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_UNCHECKED);
                }
             break;  
             case IDM_OPTIONCONTROLJOY3:
                if(use_joystick_input == 2)
                {
                   use_joystick_input = -1;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_UNCHECKED);
                } else
                {
                   use_joystick_input = 2;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_CHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_UNCHECKED);
                }
             break;  
             case IDM_OPTIONCONTROLJOY4:
                if(use_joystick_input == 3)
                {
                   use_joystick_input = -1;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_UNCHECKED);
                } else
                {
                   use_joystick_input = 3;
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY1,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY2,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY3,MF_UNCHECKED);
                   CheckMenuItem(menu,IDM_OPTIONCONTROLJOY4,MF_CHECKED);
                }
             break;                                                   
             case IDM_OPTIONSOUND:
                  if(system_sound_on==-1)
                     break;
                  if(system_sound_on)
                  {
                     FSOUND_StopSound(channel_n);
                     CheckMenuItem(menu,IDM_OPTIONSOUND,MF_UNCHECKED);
                     system_sound_on = 0;
                  }
                  else
                  {
                     channel_n = FSOUND_PlaySound(FSOUND_FREE,FSbuffer);
                     CheckMenuItem(menu,IDM_OPTIONSOUND,MF_CHECKED);
                     system_sound_on = 1;                     
                  }
             break;           
             case IDM_OPTIONSOUNDENCH1:
                if(sound_enable&SND_EN_CH1)
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH1,MF_UNCHECKED);
                   sound_enable &= ~SND_EN_CH1;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH1,MF_CHECKED);
                   sound_enable |= SND_EN_CH1;                
                }              
             break;   
             case IDM_OPTIONSOUNDENCH2:
                if(sound_enable&SND_EN_CH2)
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH2,MF_UNCHECKED);
                   sound_enable &= ~SND_EN_CH2;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH2,MF_CHECKED);
                   sound_enable |= SND_EN_CH2;                
                }              
             break;  
             case IDM_OPTIONSOUNDENCH4:
                if(sound_enable&SND_EN_CH4)
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH4,MF_UNCHECKED);
                   sound_enable &= ~SND_EN_CH4;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH4,MF_CHECKED);
                   sound_enable |= SND_EN_CH4;                
                }              
             break;               
             case IDM_OPTIONSOUNDENCH3:
                if(sound_enable&SND_EN_CH3)
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH3,MF_UNCHECKED);
                   sound_enable &= ~SND_EN_CH3;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONSOUNDENCH3,MF_CHECKED);
                   sound_enable |= SND_EN_CH3;                
                }              
             break;   
             case IDM_OPTIONSOUNDLOWPASSNONE:
                sound_lowpass = 0;
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASSNONE,MF_CHECKED);                  
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS1,MF_UNCHECKED);      
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS2,MF_UNCHECKED);                                      
             break;                     
             case IDM_OPTIONSOUNDLOWPASS1:
                sound_lowpass = LOWPASS_LEVEL1;
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS1,MF_CHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS2,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASSNONE,MF_UNCHECKED);                                                         
             break;   
             case IDM_OPTIONSOUNDLOWPASS2:
                sound_lowpass = LOWPASS_LEVEL2;
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS2,MF_CHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASS1,MF_UNCHECKED); 
                CheckMenuItem(menu,IDM_OPTIONSOUNDLOWPASSNONE,MF_UNCHECKED);                                                          
             break;         
             case IDM_OPTIONSOUNDRSTEREO:
                sound_reverse_stereo = !sound_reverse_stereo;
                CheckMenuItem(menu,IDM_OPTIONSOUNDRSTEREO,sound_reverse_stereo?MF_CHECKED:MF_UNCHECKED);                                                           
             break;      
             case IDM_OPTIONSOUNDVOL1:
                sound_volume = 1;
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL1,MF_CHECKED);   

                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL2,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL3,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL4,MF_UNCHECKED);
             break;         
             case IDM_OPTIONSOUNDVOL2:
                sound_volume = 2;
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL2,MF_CHECKED);                                                           

                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL1,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL3,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL4,MF_UNCHECKED);                
             break;       
             case IDM_OPTIONSOUNDVOL3:
                sound_volume = 3;
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL3,MF_CHECKED);

                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL1,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL2,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL4,MF_UNCHECKED);                                                                           
             break;       
             case IDM_OPTIONSOUNDVOL4:
                sound_volume = 4;
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL4,MF_CHECKED); 

                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL1,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL2,MF_UNCHECKED);  
                CheckMenuItem(menu,IDM_OPTIONSOUNDVOL3,MF_UNCHECKED);                                                                          
             break;                                                                                                                                              
             case IDM_HELPABOUT:
                  char about_str[200];
                  sprintf(about_str,str_table[ABOUT_STR],prg_version);
                  MessageBox(NULL, about_str, emu_title, 0);
             break;

           }
           break;
        case WM_INITMENU:
           FSOUND_SetMute(FSOUND_ALL,TRUE);
        break;
        case WM_EXITMENULOOP:
           if(GB1->romloaded && !paused)
              FSOUND_SetMute(FSOUND_ALL,FALSE);  
           if(sgb_mode || (sgb_border_option==SGB_INITB && border_uploaded))
              draw_border();
        break;
        case WM_ENTERSIZEMOVE:
           FSOUND_SetMute(FSOUND_ALL,TRUE); 
        break;
        case WM_EXITSIZEMOVE:
        {
           if(GB1->romloaded && !paused)   
              FSOUND_SetMute(FSOUND_ALL,FALSE); 

           POINT pt;
           GetClientRect(hwnd,&target_blt_rect);
           pt.x=pt.y=0;
           ClientToScreen(hwnd,&pt);
           OffsetRect(&target_blt_rect,pt.x,pt.y);

           if(multiple_gb) 
           { 
              if(current_gb==2)
                 target_blt_rect.left+=160*win_size;
              else 
                 target_blt_rect.right-=160*win_size; 
           }

           if(sgb_mode || (sgb_border_option==SGB_INITB && border_uploaded))
           {
              double width = ((double)(target_blt_rect.right-target_blt_rect.left)/256.0);
              double height = ((double)(target_blt_rect.bottom-target_blt_rect.top)/224.0);

              target_blt_rect.left += (long)round(48.0*width); 
              target_blt_rect.right = target_blt_rect.left + (long)round(160.0*width); 
              target_blt_rect.top += (long)round(40.0*height);
              target_blt_rect.bottom = target_blt_rect.top + (long)round(144.0*height);

              draw_border();
           }                     
           if(!GB1->romloaded || paused)
              draw_screen();           
        }
        break;
        case WM_SIZE:
        {
           POINT pt;
           GetClientRect(hwnd,&target_blt_rect);
           pt.x=pt.y=0;
           ClientToScreen(hwnd,&pt);
           OffsetRect(&target_blt_rect,pt.x,pt.y);

           if(multiple_gb) target_blt_rect.right-=160*win_size;          

           if(sgb_mode || (sgb_border_option==SGB_INITB && border_uploaded))
           {
              double width = ((double)(target_blt_rect.right-target_blt_rect.left)/256.0);
              double height = ((double)(target_blt_rect.bottom-target_blt_rect.top)/224.0);

              target_blt_rect.left += (long)round(48.0*width); 
              target_blt_rect.right = target_blt_rect.left + (long)round(160.0*width); 
              target_blt_rect.top += (long)round(40.0*height);
              target_blt_rect.bottom = target_blt_rect.top + (long)round(144.0*height);
              
              draw_border();
           }       
           if(!GB1->romloaded || paused)
              draw_screen();           
        }
        break;
        case WM_ACTIVATE:  
           if(GB1->romloaded && (sgb_mode || (sgb_border_option==SGB_INITB && border_uploaded)))
              draw_border();
           
           if(!GB1->romloaded || paused)
              draw_screen();
           
           control_pressed = 0;

           if((wParam == WA_ACTIVE) || (wParam == WA_CLICKACTIVE))
           {
              if(!menupause)
              {
                 paused = false;
                 if(GB1->romloaded)
                    FSOUND_SetMute(FSOUND_ALL,FALSE);
              }      
           } else
           {
              paused = true;
              speedup = 0;
              FSOUND_SetMute(FSOUND_ALL,TRUE);
           }
        break;
        case WM_KEYDOWN:
            switch(wParam)
            { 
            case VK_CONTROL:
               control_pressed = 1;            
            break;
            case VK_ESCAPE: // QUIT
               SendMessage(hwnd, WM_CLOSE, 0, 0L);
            break;
            case VK_F2:
               if(GB1->romloaded)
                  GB1->save_state();
            break;
            case VK_F3:
               CheckMenuItem(menu,IDM_CPUSTATESLOT0+state_slot,MF_UNCHECKED);
               
               if(++state_slot > 9)
                  state_slot = 0;
               sprintf(dx_message,"%s %d.",str_table[STATE_SLOT],state_slot);
               message_time = 40;
               
               CheckMenuItem(menu,IDM_CPUSTATESLOT0+state_slot,MF_CHECKED);                                  
            break;
            case VK_F4:
               if(GB1->romloaded)
                  GB1->load_state();
            break;
            case VK_F5:
                if(video_enable&VID_EN_BG)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENBG,MF_UNCHECKED);
                   video_enable &= ~VID_EN_BG;
                   sprintf(dx_message,"%s.","BG off");
                   message_time = 30;
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENBG,MF_CHECKED);
                   video_enable |= VID_EN_BG;                
                   sprintf(dx_message,"%s.","BG on");
                   message_time = 30;                   
                }              
            break;
            case VK_F6:
                if(video_enable&VID_EN_WIN)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENWIN,MF_UNCHECKED);
                   video_enable &= ~VID_EN_WIN;
                   sprintf(dx_message,"%s.","WIN off");
                   message_time = 30;                   
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENWIN,MF_CHECKED);
                   video_enable |= VID_EN_WIN;                
                   sprintf(dx_message,"%s.","WIN on");
                   message_time = 30;                   
                }              
            break;
            case VK_F7:
                if(video_enable&VID_EN_SPRITE)
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENSPRITE,MF_UNCHECKED);
                   video_enable &= ~VID_EN_SPRITE;
                   sprintf(dx_message,"%s.","Sprites off");
                   message_time = 30;                   
                } else
                {
                   CheckMenuItem(menu,IDM_OPTIONVIDEOENSPRITE,MF_CHECKED);
                   video_enable |= VID_EN_SPRITE;                
                   sprintf(dx_message,"%s.","Sprites on");
                   message_time = 30;                   
                }              
            break;
            #ifdef ALLOW_DEBUG
            case 'D': // DEBUG
               if(!control_pressed)
                  break;
               debug = !debug;
            break;
            case 'F': // STEP
               if(GB->romloaded && paused && !debug)
               {
                  frames = 0;
                  mainloop();
               }
               if(debug && GB->romloaded)
               {
                  frames = 1;
                  mainloop();
                  frames = 0;
                  draw_debug_screen();
               }
            break;
            #endif
            case 'P': // PAUSE
               if(!control_pressed)
                  break;
               
               if(menupause)
               {
                  CheckMenuItem(menu,IDM_CPUPAUSE,MF_UNCHECKED);
                  if(GB1->romloaded)
                     FSOUND_SetMute(FSOUND_ALL,FALSE);
               }
               else
               {
                  FSOUND_SetMute(FSOUND_ALL,TRUE);
                  CheckMenuItem(menu,IDM_CPUPAUSE,MF_CHECKED); 
               }
               menupause = !menupause;
               paused = !paused;
            break;
            case 'S': // RESET
               if(!control_pressed)
                  break;  
               soft_reset = 1;     
            break;            
            case 'R': // RESET
               if(!control_pressed)
                  break;       
                       
               if(GB1->romloaded && !write_save(1))
                  debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
                     
               reset(1);
               
               if(GB1->romloaded && !read_save(1))
                  debug_print(str_table[ERROR_SAVE_FILE_READ]);
            break;
            case '1':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[0],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[0],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[0],sizeof(unsigned int)*4);
               }
            break;
            case '2':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[1],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[1],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[1],sizeof(unsigned int)*4);
               }
            break;
            case '3':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[2],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[2],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[2],sizeof(unsigned int)*4);
               }
            break;
            case '4':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[3],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[3],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[3],sizeof(unsigned int)*4);
               }
            break;    
            case '5':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[4],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[4],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[4],sizeof(unsigned int)*4);
               }
            break; 
            case '6':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[5],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[5],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[5],sizeof(unsigned int)*4);
               }
            break; 
            case '7':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[6],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[6],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[6],sizeof(unsigned int)*4);
               }
            break; 
            case '8':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[7],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[7],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[7],sizeof(unsigned int)*4);
               }
            break; 
            case '9':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[8],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[8],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[8],sizeof(unsigned int)*4);
               }
            break; 
            case '0':
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[9],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[9],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[9],sizeof(unsigned int)*4);
               }
            break;     
            case 0xDC: 
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[10],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[10],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[10],sizeof(unsigned int)*4);
               }
            break;     
            case 0xBB: 
               if(GB->gbc_mode && !GB->rom->CGB)
               {
                  memcpy(GB->GBC_BGP,GBC_DMGBG_palettes[11],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP,GBC_DMGOBJ0_palettes[11],sizeof(unsigned int)*4);
                  memcpy(GB->GBC_OBP+4,GBC_DMGOBJ1_palettes[11],sizeof(unsigned int)*4);
               }
            break;                                                                                                 
        }
        break;
        case WM_KEYUP:
            if(wParam == VK_CONTROL)
               control_pressed = 0;
        break;
        case WM_DESTROY:            
            if(GB1->romloaded && !write_save(1))
               debug_print(str_table[ERROR_SAVE_FILE_WRITE]);

            if(GB2 && GB2->romloaded && !write_save(2))
               debug_print(str_table[ERROR_SAVE_FILE_WRITE]);            
            
            write_config_file();
            cleanup();
            PostQuitMessage(0);     
        break;
        case WM_DROPFILES:
        {
           bool romwasloaded = GB1->romloaded;
                  
           if(GB1->romloaded && !write_save(1))
              debug_print(str_table[ERROR_SAVE_FILE_WRITE]);
           HDROP d_handle = (HDROP)wParam;      
           char drag_filename[500];
           DragQueryFile(d_handle,0,drag_filename,500);
           if(!strstr(drag_filename,".zip") && !strstr(drag_filename,".ZIP") && !strstr(drag_filename,".sgb") && !strstr(drag_filename,".SGB") && !strstr(drag_filename,".GB") && !strstr(drag_filename,".gb"))
           {
              debug_print(str_table[NOT_A_ROM]);
              GB1->romloaded = false;
           } else 
              GB1->load_rom(drag_filename);
                  
           if(GB1->romloaded)
           {                    
              reset(1);
              
              if(sgb_mode)
              {
                 RECT winRect;
                 GetWindowRect(hwnd,&winRect);      
                 MoveWindow(hwnd,winRect.left,winRect.top,256*win_size+sizen_w,224*win_size+sizen_h,TRUE);
              } else
              {
                 RECT winRect;
                 GetWindowRect(hwnd,&winRect);      
                 MoveWindow(hwnd,winRect.left,winRect.top,160*win_size+sizen_w,144*win_size+sizen_h,TRUE);                     
              }
                     
              if(!read_save(1))
                 debug_print(str_table[ERROR_SAVE_FILE_READ]);
                  
              sprintf(title_text,"%s - %s",emu_title,GB1->rom_filename);
              SetWindowText(hwnd,title_text);
              if(!paused)
                 FSOUND_SetMute(FSOUND_ALL,FALSE);   
           } else 
              if(romwasloaded && GB1->cartridge != NULL) 
                 GB1->romloaded = true;    
                 
           DragFinish(d_handle);  
           SetForegroundWindow(hwnd);  
        }
        break;
        default:                  
            return DefWindowProc(hwnd, message, wParam, lParam);
    }

    return 0;
}
